/**
 * Event Writer - a utility to create string forms of events.
 *
 * $Copyright(c) 2007, 2008, 2009, 2012 Progress Software Corporation (PSC). All rights reserved.$
 * $Copyright (c) 2013, 2015, 2017 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 *
 * $Revision: 311836 $
 */
#ifndef AP_EVENTWRITER
#define AP_EVENTWRITER

#include <AP_Types.h>
#include <IAF_Platform.h>

#ifdef __cplusplus
extern "C" {
#endif  /* __cplusplus */

/**
 * Event writer types
 */
typedef enum {AP_INTEGER, AP_STRING, AP_CHAR, AP_FLOAT, AP_DECIMAL, AP_BOOLEAN, AP_SEQUENCE, AP_DICTIONARY, AP_EVENT, AP_UNFORMATTED, AP_EMPTY} AP_EventWriter_type;

struct AP_EventWriter;

/**
 * Event writer value
 */
typedef union {
	AP_int64 intValue;
	char* stringValue;
	AP_char8 charValue;
	AP_float64 floatValue;
	AP_decimal decimalValue;
	AP_bool booleanValue;
	struct AP_EventWriter* refValue;
} AP_EventWriterValue;

/**
 * Event writer element
 */
typedef struct {
	AP_EventWriterValue value;
	AP_EventWriter_type type;
} AP_EventWriterElement;


/**
 * Event writer is a utility class that can build string forms of events, 
 * sequences or dictionaries.
 *
 * An AP_EventWriter object holds a number of fields, each of which may be a
 * basic type (string, int, float, decimal, boolean), or a reference type (event, 
 * sequence, dictionary).  Once the structure has been built up by adding
 * values, the toString method can be called.  This returns a string that the
 * caller is responsible for freeing.
 *
 * Calling the destructor on an object will destroy all AP_EventWriter 
 * objects it refers to in a recursive manner - thus, only the top-level
 * object should be destroyed, and it is not possible to share AP_EventWriter
 * objects between two different containing events.
 */
struct AP_EventWriter {
	AP_EventWriter_type type;
	const char* name;
	void* keys;
	void* values;
#ifdef _DEBUG // enable duplicate dictionary key detection only in a debug build
	//shadow dictionary to prevent duplicate entries
	void* shadow_dict;
#endif // _DEBUG

	//functions
	void (*addInt)(struct AP_EventWriter* ew, AP_int64);
	void (*addString)(struct AP_EventWriter* ew, const char*);
	void (*addChar)(struct AP_EventWriter* ew, const char);
	void (*addFloat)(struct AP_EventWriter* ew, AP_float64);
	void (*addDecimal)(struct AP_EventWriter* ew, AP_decimal);
	void (*addBoolean)(struct AP_EventWriter* ew, AP_bool);
	
	void (*addRefValue)(struct AP_EventWriter* ew, struct AP_EventWriter *value);
	
	void (*addValue)(struct AP_EventWriter* ew, AP_EventWriter_type type, AP_EventWriterValue value);
	
	/** Writes a key,value pair for a dictionary. It is the caller's responsibility to 
		ensure the correct key/value types are specified, and that the same 
		key is not added to the dictionary more than once; an invalid and 
		unparseable event will result if these conditions are not met. */
	void (*addDictValue)(struct AP_EventWriter* ew, AP_EventWriter_type keytype, AP_EventWriterValue key, AP_EventWriter_type valuetype, AP_EventWriterValue value);

	/** Writes the specified value as an 'any' value. typeName must be specified
		while adding a container type (AP_SEQUENCE, AP_DICTIONARY or 'optional' type). */
	void (*addAny)(struct AP_EventWriter* ew, AP_EventWriter_type valueType, AP_EventWriterValue value, const char* typeName);
	/** Writes the specified value as an 'optional' value.*/
	void (*addOptional)(struct AP_EventWriter* ew, AP_EventWriter_type valueType, AP_EventWriterValue value);
	
	/** Returns the Apama string representation of this object. 
		The result of toString should be freed by the caller with free(). */
	char* (*toString)(struct AP_EventWriter* ew);
	void (*addUnformatted)(struct AP_EventWriter* ew, const char*);

	/** Can be called on an AP_EVENT with name as 'any' to specify its type.
		Should be called first followed by the corresponding addValue(). */
	void (*addAnyType)(struct AP_EventWriter *ew, const char*);
};

typedef struct AP_EventWriter AP_EventWriter;

/**
 * Event writer functions.
 * Those that take a string, the string need only be valid during the call (a copy is made of strings).
 * Reference types are not copied, but are destroyed by the destructor - so only need to call _dtor
 * on the top-level EventWriter object.
 * The result of toString should be freed by the caller with free().
 */
AP_IAFCORE_API AP_EventWriter* AP_IAFCORE_CALL AP_EventWriter_ctor(AP_EventWriter_type type, const char* name);
AP_IAFCORE_API void AP_IAFCORE_CALL AP_EventWriter_dtor(AP_EventWriter* ew);



#ifdef __cplusplus
}  /* extern "C" */
#endif  /* __cplusplus */

#endif
